//
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// :::::::::::::::::::::::::::::: Blobby v1.01 ::::::::::::::::::::::::::::::
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//
//
// Disclaimer: I am providing this source code "as-is," without accepting any
//             responsibility for the effects it may have on your computer,
//             network, or livelihood.  Because I am not a full time provider
//             of free Java source code, I cannot offer you any significant
//             technical support.  If you decide to use portions of my code
//             in your project or product, I would be very pleased to be 
//             attributed.  More importantly, please e-mail me to tell me
//             about it!  Do not charge money for this source code, as it is
//             intended to be freely distributed.  Do not remove this header
//             text from the source file, and do not change the version string.
//
// Author: Andrzej Kapolka, ey6es@alumni.cse.ucsc.edu
//
// Description: A common demo effect--2D metaballs.  Color and blob radius are
//              specifiable as parameters to the applet.  Compile under Java
//              version 1.1 or greater.
//
// Easy modifications: Play with the 'fps' constant to change the speed of the
//                     blobs' movement.  Modify the n-by-2 array, presetColors,
//                     to add, remove, or tweak the preset foreground/background
//                     color combinations.
//


import java.awt.*;
import java.awt.image.*;
import java.applet.*;


public class Blobby extends Applet implements Runnable {
     Thread timer = null;
     
     // Version
     String myVersion = new String("Blobby v1.01, by Andrzej Kapolka");
    
     // Timing constant
     int fps = 60;
     
     // Window size
     int winWidth, winHeight;

     // Current frame image producer and associated data
     MemoryImageSource frameSource;
     Image curFrame;
     byte[] frameData;
     int[] blobData;     

     // Palette info
     byte[] palRed;
     byte[] palGreen;
     byte[] palBlue;     

     // Color values
     Color fgColor = null, bgColor = null;
     Color[][] presetColors = { { Color.white, Color.black },
                                { Color.cyan, Color.magenta },
                                { Color.yellow, Color.blue },
                                { new Color(0.0f,1.0f,0.4f), Color.black },
                                { Color.red, Color.blue } };
     
     // Blob radius
     int radius = 64;
           
           
     public void init() {
          int i, j, dx, dy, blobWidth, blobHeight;
          float v, vc, thresh, scale, dist;

          if(getParameter("fgColor")!=null) {
              fgColor = Color.decode(getParameter("fgColor"));          
          }

          if(getParameter("bgColor")!=null) {
              bgColor = Color.decode(getParameter("bgColor"));          
          }
         
          if(getParameter("blobRadius")!=null) {
              radius = Integer.parseInt(getParameter("blobRadius"));
          }
         
          
          // Choose a random color combination if none specified
          
          i = (int)(Math.random()*presetColors.length);
          
          if(fgColor==null) {
               fgColor = presetColors[i][0];
          }
      
          if(bgColor==null) {
               bgColor = presetColors[i][1];
          }


          winWidth = size().width;
          winHeight = size().height;

          palRed = new byte[256];
          palGreen = new byte[256];
          palBlue = new byte[256];

          buildPalette(0.0f);


          frameData = new byte[winWidth*winHeight];
          frameSource = new MemoryImageSource(winWidth,winHeight,
                                              new IndexColorModel(8,256,
                                                                  palRed,palGreen,
                                                                  palBlue),
                                              frameData,0,winWidth);
          frameSource.setAnimated(true);
          curFrame = createImage(frameSource);


          // Make blob thing

          blobWidth = winWidth*2;
          blobHeight = winHeight*2;

          blobData = new int[blobWidth*blobHeight];

          thresh = (float)radius*radius;
          scale = 255.0f*thresh;
          
          for(i=0;i<blobHeight;i++) {
               for(j=0;j<blobWidth;j++) {
                    dx = (j-(blobWidth>>1));
                    dy = (i-(blobHeight>>1));

                    if((dist=dx*dx+dy*dy)<thresh) dist = thresh;

                    blobData[i*blobWidth+j] = (int)(scale/dist);
               }
          }
     }

     public void buildPalette(float t) {
          int i;
          float u = (1.0f-t), v, vc;
          
          for(i=0;i<256;i++) {
          	v = (i/255.0f)*u+(1.0f-(i/255.0f))*t;
          	vc = 1.0f-v;
          	
               palRed[i] = (byte)(bgColor.getRed()*vc+fgColor.getRed()*v);
               palGreen[i] = (byte)(bgColor.getGreen()*vc+fgColor.getGreen()*v);
               palBlue[i] = (byte)(bgColor.getBlue()*vc+fgColor.getBlue()*v);
          }
     }

     public void start() {
          
          // Start the timer thread
          timer = new Thread(this);
          timer.start();
     }

     public void stop() {
          timer = null;
     }
     
     public void run() {
          int i, j, tmp, dptr, blobWidth = (winWidth<<1), countdown = 330;
          double xTemp, yTemp, len, xCent, yCent;
          boolean origColor = true;
          
          int[] sptr1 = new int[3], sptr2 = new int[3];
          double[] xOff = new double[3], yOff = new double[3],
                   xVel = new double[3], yVel = new double[3];

          for(i=0;i<3;i++) {
               xOff[i] = Math.random()*winWidth;
               yOff[i] = Math.random()*winHeight;

               xVel[i] = (Math.random()*4.0)-2.0;
               yVel[i] = (Math.random()*4.0)-2.0;
          }

          showStatus(myVersion);
          
          while(timer!=null) {
               
               xCent = (xOff[0]+xOff[1]+xOff[2])/3.0;
               yCent = (yOff[0]+yOff[1]+yOff[2])/3.0;
               
               for(i=0;i<3;i++) {
                    xOff[i]+=xVel[i];
                    yOff[i]+=yVel[i];

                    if(xOff[i]>=winWidth || xOff[i]<0.0) 
                         { xOff[i] -= xVel[i]; xVel[i] *= -1.0; }

                    if(yOff[i]>=winHeight || yOff[i]<0.0) 
                         { yOff[i] -= yVel[i]; yVel[i] *= -1.0; }
                    
                    // Move towards center, usually--but repel once in a while
                    
                    xTemp = xCent-xOff[i]; yTemp = yCent-yOff[i];
                    len = Math.sqrt(xTemp*xTemp+yTemp*yTemp);
                    xTemp /= len; yTemp /= len;
                    
                    if(Math.random()<0.02) {
                         len=Math.random()*8.0;
                         
                         xVel[i]-=xTemp*len; 
                         yVel[i]-=yTemp*len;
                    }
                    else { 
                         xVel[i]+=xTemp; 
                         yVel[i]+=yTemp;
                    }
                    
                    
                    sptr1[i] = ((int)yOff[i])*blobWidth+(int)xOff[i];
               }

               dptr = 0;
               

               for(i=0;i<winHeight;i++) {
                    sptr2[0] = sptr1[0];
                    sptr2[1] = sptr1[1];
                    sptr2[2] = sptr1[2];

                    for(j=0;j<winWidth;j++) {
                         tmp = ((blobData[sptr2[0]++])+
                                (blobData[sptr2[1]++])+
                                (blobData[sptr2[2]++]));

                         if(tmp<=0) frameData[dptr++] = (byte)0;
                         else if(tmp>=255) frameData[dptr++] = (byte)255;
                         else frameData[dptr++] = (byte)tmp;
                    }

                    sptr1[0]+=blobWidth;
                    sptr1[1]+=blobWidth;
                    sptr1[2]+=blobWidth;
               }

               if((countdown--)<=30) {
                    if(origColor) buildPalette(1.0f-(countdown/30.0f));
                    else buildPalette(countdown/30.0f);
                    
                    frameSource.newPixels(frameData,
                                          new IndexColorModel(8,256,palRed,
                                                              palGreen,palBlue),
                                          0,winWidth);
                                     
                    if(countdown==0) {
                         countdown=330;
                         origColor = !origColor;
                         
                         showStatus(myVersion);
                    }
               }
               else frameSource.newPixels();
               
               repaint();

               try { Thread.sleep(1000/fps); }
               catch(InterruptedException e) { }
          }
     }

     public void update(Graphics g) {
          g.drawImage(curFrame,0,0,this);
     }
}